from inspect import isawaitable
from typing import List, Optional, Protocol

from graphene import Context, ResolveInfo
from graphql import GraphQLError


class Extension(Protocol):
    def __init__(self):
        pass  # pragma: no cover

    def request_started(self, context: Context):
        pass  # pragma: no cover

    def request_finished(self, context: Context):
        pass  # pragma: no cover

    async def resolve(self, next_, parent, info: ResolveInfo, **kwargs):
        result = next_(parent, info, **kwargs)
        if isawaitable(result):
            result = await result
        return result

    def has_errors(self, errors: List[GraphQLError], context: Context):
        pass  # pragma: no cover

    def format(self, context: Context) -> Optional[dict]:
        pass  # pragma: no cover



from contextlib import contextmanager
from typing import Any, Dict, List, Type

from graphene import Context
from graphql import GraphQLError

from .extension import Extension


class ExtensionManager:
    __slots__ = ("context", "extensions")

    def __init__(
        self,
        extensions: List[Type[Extension]] = None,
        context: Context = None,
    ):
        self.context = context

        if extensions:
            self.extensions = tuple(ext() for ext in extensions)
        else:
            self.extensions = ()

    @contextmanager
    def request(self):
        for ext in self.extensions:
            ext.request_started(self.context)
        try:
            yield
        finally:
            for ext in self.extensions[::-1]:
                ext.request_finished(self.context)

    def has_errors(self, errors: List[GraphQLError]):
        for ext in self.extensions:
            ext.has_errors(errors, self.context)

    def format(self) -> Dict[str, Any]:
        data = {}
        for ext in self.extensions:
            ext_data = ext.format(self.context)
            if ext_data:
                data.update(ext_data)
        return data



from contextlib import asynccontextmanager
from inspect import isawaitable
from typing import Any, Awaitable, Callable, Dict, List, Optional, Type, Union

from graphene import Context
from graphql import ExecutionResult, graphql
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession
from starlette.background import BackgroundTasks
from starlette.requests import HTTPConnection, Request
from starlette.responses import JSONResponse, Response
from starlette_graphene3 import (
    GraphQLApp,
    _get_operation_from_request,
    make_graphiql_handler,
)

from .extensions import Extension, ExtensionManager

DEFAULT_GET = object()


class SessionQLApp(GraphQLApp):
    def __init__(
        self,
        engine: AsyncEngine,
        context_value: Callable = Context,
        on_get: Optional[
            Callable[[Request], Union[Response, Awaitable[Response]]]
        ] = DEFAULT_GET,
        extensions: List[Type[Extension]] = (),
        *args,
        **kwargs,
    ):
        self.engine = engine
        if on_get == DEFAULT_GET:
            on_get = make_graphiql_handler()

        self.extensions = extensions or ()
        super().__init__(context_value=context_value, on_get=on_get, *args, **kwargs)

    async def _handle_http_request(self, request: Request) -> JSONResponse:
        try:
            operations = await _get_operation_from_request(request)
        except ValueError as e:
            return JSONResponse({"errors": [e.args[0]]}, status_code=400)

        if isinstance(operations, list):
            return JSONResponse(
                {"errors": ["This server does not support batching"]}, status_code=400
            )
        else:
            operation = operations

        query = operation["query"]
        variable_values = operation.get("variables")
        operation_name = operation.get("operationName")

        async with self._get_context_value(request) as context_value:
            middleware = self.middleware or ()
            extension_manager = ExtensionManager(self.extensions, context=context_value)

            with extension_manager.request():
                result: ExecutionResult = await graphql(
                    self.schema.graphql_schema,
                    source=query,
                    context_value=context_value,
                    root_value=self.root_value,
                    middleware=(*middleware, *extension_manager.extensions),
                    variable_values=variable_values,
                    operation_name=operation_name,
                    execution_context_class=self.execution_context_class,
                )

            extension_results = extension_manager.format()
            if extension_results:
                result.extensions = extension_results

            background = getattr(context_value, "background", None)

        response: Dict[str, Any] = {"data": result.data}
        if result.errors:
            for error in result.errors:
                if error.original_error:
                    self.logger.error(
                        "An exception occurred in resolvers",
                        exc_info=error.original_error,
                    )
            response["errors"] = [
                self.error_formatter(error) for error in result.errors
            ]
        if result.extensions:
            response["extensions"] = result.extensions

        return JSONResponse(
            response,
            status_code=200,
            background=background,
        )

    @asynccontextmanager
    async def _get_context_value(self, request: HTTPConnection) -> Context:
        async with AsyncSession(self.engine) as session:
            async with session.begin():
                if callable(self.context_value):
                    context = self.context_value(
                        request=request,
                        background=BackgroundTasks(),
                        session=session,
                    )
                    if isawaitable(context):
                        context = await context
                    yield context
                else:
                    yield self.context_value or Context(
                        request=request,
                        background=BackgroundTasks(),
                        session=session,
                    )



import hashlib
import json
from typing import List, Type

import graphene
import sqlalchemy as sa
from graphene.types.enum import EnumMeta
from graphql import StringValueNode
from sqlalchemy.orm import DeclarativeMeta

from .gql_id import ResolvedGlobalId
from .registry import get_global_registry
from .sqlalchemy_converter import convert_sqlalchemy_type

# There we contain unique type names
initialized_types = {}


def get_unique_input_type_name(
    model_name: str, input_fields: dict, operation: str
) -> str:
    input_type_name = f"Input{operation}{model_name}"
    model_hash = hashlib.md5(
        json.dumps(
            {
                key: f"{type(value)}{getattr(value, 'kwargs', '') or ''}"
                for key, value in input_fields.items()
            },
            sort_keys=True,
        ).encode("utf-8")
    ).hexdigest()

    if initialized_types.setdefault(input_type_name, model_hash) == model_hash:
        return input_type_name

    input_type_name = input_type_name + "_"
    for i in model_hash:
        input_type_name += i
        if initialized_types.setdefault(input_type_name, model_hash) == model_hash:
            return input_type_name


class ArgID(graphene.Scalar):
    @staticmethod
    def coerce_id(value):
        global_id = ResolvedGlobalId.decode(value)

        # registry = get_global_registry()
        # type_ = registry.get_type_for_model(table)
        # if type_ and type_.__name__ != global_id.type:
        #     raise Exception(
        #         f"Invalid GlobalID type: {global_id.type} != {type_.__name__}"
        #     )

        return global_id.id

    serialize = coerce_id
    parse_value = coerce_id

    @staticmethod
    def parse_literal(ast):
        if isinstance(ast, StringValueNode):
            return ast.value


def convert_sqlalchemy_type_mutation(column):
    field = convert_sqlalchemy_type(
        getattr(column, "type", None),
        column,
    )
    if field == graphene.Int and column.foreign_keys:
        field = ArgID()

    return field


def get_input_fields(
    model: Type[DeclarativeMeta],
    only_fields: List = (),
    exclude_fields: List = (),
    required_fields: List = (),
) -> dict:
    if only_fields and exclude_fields:
        raise ValueError(
            "The options 'only_fields' and 'exclude_fields' cannot be both set on the same type."
        )

    table = sa.inspect(model).mapped_table

    fields = {}
    for name, column in dict(table.columns).items():
        if only_fields and name not in only_fields:
            continue
        if exclude_fields and name in exclude_fields:
            continue

        field = convert_sqlalchemy_type_mutation(column)

        if callable(field):
            field = field()
        if isinstance(field, EnumMeta):
            field = field()

        if name in required_fields:
            getattr(field, "kwargs", {})["required"] = True

        fields[name] = field

    return fields


def get_input_type(
    model: Type[DeclarativeMeta], input_fields: dict, operation: str
) -> Type:
    return type(
        get_unique_input_type_name(
            model_name=model.__name__, input_fields=input_fields, operation=operation
        ),
        (graphene.InputObjectType,),
        input_fields,
    )



from functools import partial
from inspect import isawaitable

from graphene.types import Field, ID, Interface
from graphene.types.interface import InterfaceOptions
from graphene.types.utils import get_type

from .gql_id import ResolvedGlobalId


class AsyncGlobalID(Field):
    def __init__(self, node=None, parent_type=None, required=True, *args, **kwargs):
        super().__init__(ID, required=required, *args, **kwargs)
        self.node = node or AsyncNode
        self.parent_type_name = parent_type._meta.name if parent_type else None

    @staticmethod
    async def id_resolver(
        parent_resolver, node, root, info, parent_type_name=None, **args
    ):
        type_id = parent_resolver(root, info, **args)
        if isawaitable(type_id):
            type_id = await type_id
        parent_type_name = parent_type_name or info.parent_type.name
        return node.to_global_id(parent_type_name, type_id)  # root._meta.name

    def wrap_resolve(self, parent_resolver):
        return partial(
            self.id_resolver,
            parent_resolver,
            self.node,
            parent_type_name=self.parent_type_name,
        )


class AsyncNodeField(Field):
    def __init__(self, node, type_=False, **kwargs):
        assert issubclass(node, AsyncNode), "NodeField can only operate in Nodes"
        self.node_type = node
        self.field_type = type_

        super().__init__(
            # If we don's specify a type, the field type will be the node
            # interface
            type_ or node,
            id=ID(required=True, description="The ID of the object"),
            **kwargs,
        )

    def wrap_resolve(self, parent_resolver):
        return partial(self.node_type.node_resolver, get_type(self.field_type))


class AbstractAsyncNode(Interface):
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(cls, **options):
        _meta = InterfaceOptions(cls)
        _meta.fields = {"id": AsyncGlobalID(cls, description="The ID of the object")}
        super().__init_subclass_with_meta__(_meta=_meta, **options)


class AsyncNode(AbstractAsyncNode):
    """An object with an ID"""

    @classmethod
    def Field(cls, *args, **kwargs):  # noqa: N802
        return AsyncNodeField(cls, *args, **kwargs)

    @classmethod
    def node_resolver(cls, only_type, root, info, id):
        return cls.get_node_from_global_id(info, id, only_type=only_type)

    @classmethod
    def get_node_from_global_id(cls, info, global_id, only_type=None):
        try:
            _type, _id = cls.from_global_id(global_id)
        except Exception as e:
            raise Exception(
                (
                    f'Unable to parse global ID "{global_id}". '
                    'Make sure it is a base64 encoded string in the format: "TypeName:id". '
                    f"Exception message: {str(e)}"
                )
            )

        graphene_type = info.schema.get_type(_type)
        if graphene_type is None:
            raise Exception(f'Relay Node "{_type}" not found in schema')

        graphene_type = graphene_type.graphene_type

        if only_type:
            assert (
                graphene_type == only_type
            ), f"Must receive a {only_type._meta.name} id."

        # We make sure the ObjectType implements the "Node" interface
        if cls not in graphene_type._meta.interfaces:
            raise Exception(
                f'ObjectType "{_type}" does not implement the "{cls}" interface.'
            )

        get_node = getattr(graphene_type, "get_node", None)
        if get_node:
            return get_node(info, _id)

    @classmethod
    def from_global_id(cls, global_id) -> ResolvedGlobalId:
        return ResolvedGlobalId.decode(global_id)

    @classmethod
    def to_global_id(cls, type_, id):
        return ResolvedGlobalId(type_, id).encode()



import binascii
from base64 import b64decode, b64encode
from typing import Optional

PREFIX = "arrayconnection:"


def base64(s: str) -> str:
    """Encode the string s using Base64."""
    b: bytes = s.encode("utf-8") if isinstance(s, str) else s
    return b64encode(b).decode("ascii")


def unbase64(s: str) -> str:
    """Decode the string s using Base64."""
    try:
        b: bytes = s.encode("ascii") if isinstance(s, str) else s
    except UnicodeEncodeError:
        return ""
    try:
        return b64decode(b).decode("utf-8")
    except (binascii.Error, UnicodeDecodeError):
        return ""


def offset_to_cursor(offset: int) -> str:
    """Create the cursor string from an offset."""
    return base64(f"{PREFIX}{offset}")


def cursor_to_offset(cursor: str) -> Optional[int]:
    """Extract the offset from the cursor string."""
    try:
        return int(unbase64(cursor)[len(PREFIX) :])
    except ValueError:
        return None


def get_offset_with_default(
    cursor: Optional[str] = None, default_offset: int = 0
) -> int:
    """Get offset from a given cursor and a default.

    Given an optional cursor and a default offset, return the offset to use;
    if the cursor contains a valid offset, that will be used,
    otherwise it will be the default.
    """
    if not isinstance(cursor, str):
        return default_offset

    offset = cursor_to_offset(cursor)
    return default_offset if offset is None else offset



import graphene
import pytest
from graphene import Context

from alchql.fields import SQLAlchemyConnectionField
from alchql.middlewares import LoaderMiddleware
from alchql.node import AsyncNode
from alchql.sql_mutation import (
    SQLAlchemyDeleteMutation,
)
from alchql.types import SQLAlchemyObjectType
from tests.models import Pet
from tests.test_query import add_test_data


@pytest.mark.asyncio
async def test_delete_mutation(session):
    await add_test_data(session)

    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            interfaces = (AsyncNode,)

    class MutationDeletePet(SQLAlchemyDeleteMutation):
        class Meta:
            model = Pet
            output = PetType

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        all_pets = SQLAlchemyConnectionField(PetType.connection)

    class Mutation(graphene.ObjectType):
        delete_pet = MutationDeletePet.Field()

    query = """
            query {
                allPets {
                    edges {
                        node {
                            id
                            name
                        }
                    }
                }
            }
        """

    schema = graphene.Schema(
        query=Query,
        mutation=Mutation,
    )
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Pet]),
        ],
    )
    assert not result.errors

    all_pets = [i["node"] for i in result.data["allPets"]["edges"]]

    assert [i["name"] for i in all_pets] == [
        "Garfield",
        "Lassie",
    ]

    query = """
        mutation UpdatePet($updatePetId: ID!) {
            deletePet(id: $updatePetId) {
                id
                name
            }
        }
    """

    id_to_update = all_pets[0]["id"]
    new_name = "bear"

    result = await schema.execute_async(
        query,
        variables={
            "updatePetId": id_to_update,
        },
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Pet]),
        ],
    )

    assert not result.errors
    # TODO: problems with sqlite
    assert result.data["deletePet"] is None



import enum
from collections import defaultdict

import sqlalchemy as sa
from aiodataloader import DataLoader
from graphene import ResolveInfo
from sqlalchemy import ForeignKey, Table
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import RelationshipProperty

from .query_helper import QueryHelper
from .utils import EnumValue, filter_requested_fields_for_object, table_to_class


def generate_loader_by_relationship(relation: RelationshipProperty):
    class Loader(DataLoader):
        def __init__(
            self,
            session: AsyncSession,
            info: ResolveInfo = None,
            *args,
            **kwargs,
        ):
            self.session = session
            self.info = info
            self.fields = set()
            super().__init__(*args, **kwargs)

        async def batch_load_fn(self, keys):
            f = next(iter(relation.local_columns))
            target = relation.mapper.entity

            object_types = getattr(self.info.context, "object_types", {})
            setattr(self.info.context, "keys", keys)
            object_type = object_types.get(self.info.field_name)

            filters = QueryHelper.get_filters(self.info)
            gql_field = QueryHelper.get_current_field(self.info)
            sort_args = []
            sort = []
            if (
                "sort" in gql_field.arguments
                and gql_field.arguments["sort"] is not None
            ):
                sort = gql_field.arguments["sort"]
                if not isinstance(sort, list):
                    sort = [sort]

                if hasattr(object_type, "sort_argument"):
                    sort_type = object_type.sort_argument().type.of_type
                    new_sort = []
                    for s in sort:
                        new_sort.append(getattr(sort_type, s).value)

                    sort = new_sort

                sort_args = []
                # ensure consistent handling of graphene Enums, enum values and
                # plain strings
                for item in sort:
                    if isinstance(item, (EnumValue, enum.Enum)):
                        sort_args.append(item.value.nullslast())
                    else:
                        sort_args.append(item.nullslast())

            selected_fields = QueryHelper.get_selected_fields(
                self.info, model=target, sort=sort
            )
            if not selected_fields:
                selected_fields = self.fields or target.__table__.columns

            q = sa.select(
                *selected_fields,
                f.label("_batch_key"),
            )

            if relation.primaryjoin is not None:
                q = q.where(relation.primaryjoin)
            if relation.secondaryjoin is not None:
                q = q.where(relation.secondaryjoin)

            if relation.order_by:
                for ob in relation.order_by:
                    q = q.order_by(ob.nullslast())

            q = q.where(f.in_(keys))

            if object_type and hasattr(object_type, "set_select_from"):
                setattr(self.info.context, "keys", keys)
                q = await object_type.set_select_from(self.info, q, gql_field.values)
                if list(q._group_by_clause):
                    q = q.group_by(f)

            if filters:
                q = q.where(sa.and_(*filters))

            q = q.order_by(*sort_args)

            results_by_ids = defaultdict(list)

            conversion_type = object_type or target
            for result in await self.session.execute(q.distinct()):
                _data = dict(**result)
                _batch_key = _data.pop("_batch_key")
                _data = filter_requested_fields_for_object(_data, conversion_type)
                results_by_ids[_batch_key].append(conversion_type(**_data))

            return [results_by_ids.get(id, []) for id in keys]

    return Loader


def generate_loader_by_foreign_key(fk: ForeignKey, reverse=False):
    class Loader(DataLoader):
        def __init__(
            self,
            session: AsyncSession,
            info: ResolveInfo = None,
            *args,
            **kwargs,
        ):
            self.session = session
            self.info = info
            self.fields = set()
            super().__init__(*args, **kwargs)

        async def batch_load_fn(self, keys):
            if not reverse:
                target_field = fk.column
                source_field = fk.parent
                target: Table = target_field.table
                source: Table = source_field.table
            else:
                target_field = fk.parent
                source_field = fk.column
                target: Table = target_field.table
                source: Table = source_field.table

            object_types = getattr(self.info.context, "object_types", {})
            setattr(self.info.context, "keys", keys)
            object_type = object_types.get(self.info.field_name)

            filters = QueryHelper.get_filters(self.info)
            gql_field = QueryHelper.get_current_field(self.info)
            sort_args = []
            sort = []
            if (
                "sort" in gql_field.arguments
                and gql_field.arguments["sort"] is not None
            ):
                sort = gql_field.arguments["sort"]
                if not isinstance(sort, list):
                    sort = [sort]

                if hasattr(object_type, "sort_argument"):
                    sort_type = object_type.sort_argument().type.of_type
                    new_sort = []
                    for s in sort:
                        new_sort.append(getattr(sort_type, s).value)

                    sort = new_sort

                sort_args = []
                # ensure consistent handling of graphene Enums, enum values and
                # plain strings
                for item in sort:
                    if isinstance(item, (EnumValue, enum.Enum)):
                        sort_args.append(item.value.nullslast())
                    else:
                        sort_args.append(item.nullslast())

            selected_fields = QueryHelper.get_selected_fields(
                self.info, model=target, sort=sort
            )

            if not selected_fields:
                selected_fields = self.fields or target.columns

            q = sa.select(
                *selected_fields,
                target_field.label("_batch_key"),
            ).where(target_field.in_(keys))

            if target != source:
                q = q.select_from(
                    sa.outerjoin(
                        target,
                        source,
                        target_field == source_field,
                    )
                )

            if object_type and hasattr(object_type, "set_select_from"):
                setattr(self.info.context, "keys", keys)
                q = await object_type.set_select_from(self.info, q, gql_field.values)
                if list(q._group_by_clause):
                    q = q.group_by(source_field)

            if filters:
                q = q.where(sa.and_(*filters))

            q = q.order_by(*sort_args)

            results_by_ids = defaultdict(list)

            conversion_type = object_type or table_to_class(target)
            results = list(map(dict, await self.session.execute(q.distinct())))
            for _data in results:
                _batch_key = _data.pop("_batch_key")
                _data = filter_requested_fields_for_object(_data, conversion_type)
                results_by_ids[_batch_key].append(conversion_type(**_data))

            return [results_by_ids.get(id, []) for id in keys]

    return Loader



import re
from collections import OrderedDict
from typing import Callable, Optional, Tuple, Type

import sqlalchemy
from graphene import Field, ResolveInfo
from graphene.relay import Connection, Node
from graphene.types.objecttype import ObjectType, ObjectTypeOptions
from graphene.types.utils import yank_fields_from_attrs
from graphene.utils.orderedtype import OrderedType
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import (
    ColumnProperty,
    CompositeProperty,
    DeclarativeMeta,
    RelationshipProperty,
)

from .converter import (
    convert_sqlalchemy_column,
    convert_sqlalchemy_composite,
    convert_sqlalchemy_fk,
    convert_sqlalchemy_fk_reverse,
    convert_sqlalchemy_hybrid_method,
    convert_sqlalchemy_relationship,
)
from .enums import (
    enum_for_field,
    sort_argument_for_object_type,
    sort_enum_for_object_type,
)
from .node import AsyncNode
from .registry import Registry, get_global_registry
from .resolvers import get_attr_resolver, get_custom_resolver
from .utils import get_query, is_mapped_class, is_mapped_instance


class ORMField(OrderedType):
    def __init__(
        self,
        model_attr: str = None,
        type_=None,
        required: bool = None,
        description: str = None,
        deprecation_reason: str = None,
        batching: bool = None,
        _creation_counter: int = None,
        **field_kwargs,
    ):
        """
        Use this to override fields automatically generated by SQLAlchemyObjectType.
        Unless specified, options will default to SQLAlchemyObjectType usual behavior
        for the given SQLAlchemy model property.

        Usage:
            class MyModel(Base):
                id = Column(Integer(), primary_key=True)
                name = Column(String)

            class MyType(SQLAlchemyObjectType):
                class Meta:
                    model = MyModel

                id = ORMField(type=graphene.Int)
                name = ORMField(required=True)

        -> MyType.id will be of type Int (vs ID).
        -> MyType.name will be of type NonNull(String) (vs String).

        :param str model_attr:
            Name of the SQLAlchemy model attribute used to resolve this field.
            Default to the name of the attribute referencing the ORMField.
        :param type_:
            Default to the type mapping in converter.py.
        :param str description:
            Default to the `doc` attribute of the SQLAlchemy column property.
        :param bool required:
            Default to the opposite of the `nullable` attribute of the SQLAlchemy column property.
        :param str description:
            Same behavior as in graphene.Field. Defaults to None.
        :param str deprecation_reason:
            Same behavior as in graphene.Field. Defaults to None.
        :param bool batching:
            Toggle SQL batching. Defaults to None, that is `SQLAlchemyObjectType.meta.batching`.
        :param int _creation_counter:
            Same behavior as in graphene.Field.
        """
        super().__init__(_creation_counter=_creation_counter)
        # The is only useful for documentation and auto-completion
        common_kwargs = {
            "model_attr": model_attr,
            "type_": type_,
            "required": required,
            "description": description,
            "deprecation_reason": deprecation_reason,
            "batching": batching,
        }
        common_kwargs = {
            kwarg: value for kwarg, value in common_kwargs.items() if value is not None
        }
        self.kwargs = field_kwargs
        self.kwargs.update(common_kwargs)


def construct_fields(
    obj_type: Type["SQLAlchemyObjectType"],
    model: DeclarativeMeta,
    registry: Registry,
    only_fields: Tuple[str, ...],
    exclude_fields: Tuple[str, ...],
    connection_field_factory: Optional[Callable],
) -> OrderedDict[str, Field]:
    """
    Construct all the fields for a SQLAlchemyObjectType.
    The main steps are:
      - Gather all the relevant attributes from the SQLAlchemy model
      - Gather all the ORM fields defined on the type
      - Merge in overrides and build up all the fields
    """
    inspected_model = sqlalchemy.inspect(model)
    # Gather all the relevant attributes from the SQLAlchemy model in order
    all_model_attrs = OrderedDict(
        inspected_model.column_attrs.items()
        + inspected_model.composites.items()
        + [
            (name, item)
            for name, item in inspected_model.all_orm_descriptors.items()
            if isinstance(item, hybrid_property)
        ]
        + inspected_model.relationships.items()
    )

    # Filter out excluded fields
    auto_orm_field_names = []
    for attr_name, attr in all_model_attrs.items():
        if (only_fields and attr_name not in only_fields) or (
            attr_name in exclude_fields
        ):
            continue
        auto_orm_field_names.append(attr_name)

    # Gather all the ORM fields defined on the type
    custom_orm_fields_items = [
        (attn_name, attr)
        for base in reversed(obj_type.__mro__)
        for attn_name, attr in base.__dict__.items()
        if isinstance(attr, ORMField)
    ]
    custom_orm_fields_items = sorted(custom_orm_fields_items, key=lambda item: item[1])

    # Set the model_attr if not set
    for orm_field_name, orm_field in custom_orm_fields_items:
        attr_name = orm_field.kwargs.get("model_attr", orm_field_name)
        if attr_name not in all_model_attrs:
            raise ValueError(
                (
                    f"Cannot map ORMField to a model attribute.\n"
                    f"Field: '{obj_type.__name__}.{orm_field_name}'"
                )
            )
        orm_field.kwargs["model_attr"] = attr_name

    # Merge automatic fields with custom ORM fields
    orm_fields = OrderedDict(custom_orm_fields_items)
    for orm_field_name in auto_orm_field_names:
        if orm_field_name in orm_fields:
            continue
        orm_fields[orm_field_name] = ORMField(model_attr=orm_field_name)

    # Build all the field dictionary
    auto_fields = OrderedDict()
    for mapper in inspected_model.registry.mappers:
        for fk in mapper.local_table.foreign_keys:
            if inspected_model.selectable == fk.column.table:
                orm_field_name = str(fk.parent.table.fullname)
                if (only_fields and orm_field_name not in only_fields) or (
                    orm_field_name in exclude_fields
                ):
                    continue
                auto_fields[orm_field_name] = convert_sqlalchemy_fk_reverse(
                    fk, obj_type
                )

    for fk in inspected_model.mapped_table.foreign_keys:
        orm_field_name = re.sub(r"_(?:id|pk)$", "", fk.parent.key)
        if (only_fields and orm_field_name not in only_fields) or (
            orm_field_name in exclude_fields
        ):
            continue
        auto_fields[orm_field_name] = convert_sqlalchemy_fk(
            fk,
            obj_type,
            orm_field_name,
        )

    fields = OrderedDict()
    for orm_field_name, orm_field in orm_fields.items():
        attr_name = orm_field.kwargs.pop("model_attr")
        attr = all_model_attrs[attr_name]
        resolver = get_custom_resolver(obj_type, orm_field_name) or get_attr_resolver(
            attr_name
        )

        if isinstance(attr, ColumnProperty):
            field = convert_sqlalchemy_column(
                attr, registry, resolver, **orm_field.kwargs
            )
        elif isinstance(attr, RelationshipProperty):
            field = convert_sqlalchemy_relationship(
                attr,
                obj_type,
                connection_field_factory,
                orm_field_name,
                **orm_field.kwargs,
            )
        elif isinstance(attr, CompositeProperty):
            if attr_name != orm_field_name or orm_field.kwargs:
                # TODO Add a way to override composite property fields
                raise ValueError(
                    "ORMField kwargs for composite fields must be empty. "
                    f"Field: {obj_type.__name__}.{orm_field_name}"
                )
            field = convert_sqlalchemy_composite(attr, registry, resolver)
        elif isinstance(attr, hybrid_property):
            field = convert_sqlalchemy_hybrid_method(attr, resolver, **orm_field.kwargs)
        else:
            raise Exception("Property type is not supported")  # Should never happen

        registry.register_orm_field(obj_type, orm_field_name, attr)
        fields[orm_field_name] = field

    return OrderedDict(auto_fields | fields)


class SQLAlchemyObjectTypeOptions(ObjectTypeOptions):
    model: Type[DeclarativeMeta] = None
    registry: Registry = None
    connection: Type[Connection] = None
    id: str = None


class SQLAlchemyObjectType(ObjectType):
    _meta: SQLAlchemyObjectTypeOptions

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        model: DeclarativeMeta = None,
        registry: Registry = None,
        skip_registry=False,
        only_fields=(),
        exclude_fields=(),
        connection=None,
        connection_class=None,
        use_connection=None,
        interfaces=(),
        id=None,
        connection_field_factory=None,
        _meta=None,
        **options,
    ):
        assert is_mapped_class(model), (
            "You need to pass a valid SQLAlchemy Model in "
            f'{cls.__name__}.Meta, received "{model}".'
        )

        if not registry:
            registry = get_global_registry()

        assert isinstance(registry, Registry), (
            f"The attribute registry in {cls.__name__} needs to be an instance of "
            f'Registry, received "{registry}".'
        )

        if only_fields and exclude_fields:
            raise ValueError(
                "The options 'only_fields' and 'exclude_fields' cannot be both set on the same type."
            )

        sqla_fields = yank_fields_from_attrs(
            construct_fields(
                obj_type=cls,
                model=model,
                registry=registry,
                only_fields=only_fields,
                exclude_fields=exclude_fields,
                connection_field_factory=connection_field_factory,
            ),
            _as=Field,
            sort=False,
        )

        if use_connection is None and interfaces:
            use_connection = any(
                issubclass(interface, (Node, AsyncNode)) for interface in interfaces
            )

        if use_connection and not connection:
            # We create the connection automatically
            if not connection_class:
                connection_class = Connection

            connection = connection_class.create_type(
                f"{cls.__name__}Connection", node=cls
            )

        if connection is not None:
            assert issubclass(
                connection, Connection
            ), f"The connection must be a Connection. Received {connection.__name__}"

        if not _meta:
            _meta = SQLAlchemyObjectTypeOptions(cls)

        _meta.model = model
        _meta.registry = registry

        if _meta.fields:
            _meta.fields.update(sqla_fields)
        else:
            _meta.fields = sqla_fields

        _meta.connection = connection
        _meta.id = id or "id"

        if options.get("filter_fields"):
            _meta.filter_fields = options["filter_fields"]

        cls.connection = connection  # Public way to get the connection

        super().__init_subclass_with_meta__(
            _meta=_meta, interfaces=interfaces, **options
        )

        if not skip_registry:
            registry.register(cls)

    @classmethod
    def is_type_of(cls, root, info: ResolveInfo):
        if isinstance(root, cls):
            return True
        if not is_mapped_instance(root):
            raise Exception(f'Received incompatible instance "{root}".')
        return isinstance(root, cls._meta.model)

    @classmethod
    async def get_query(cls, info: ResolveInfo):
        return get_query(cls._meta.model, info, cls.__name__)

    @classmethod
    async def get_node(cls, info: ResolveInfo, id):
        session = info.context.session

        pk = sqlalchemy.inspect(cls._meta.model).primary_key[0]
        q = (await cls.get_query(info)).where(pk == id)
        obj = (await session.execute(q)).first()
        if obj:
            result = cls(**obj)
            return result

    async def resolve_id(self, info: ResolveInfo):
        key = "id"
        if isinstance(self, SQLAlchemyObjectType):
            model = self._meta.model
            key = model.__mapper__.primary_key[0].key
        return getattr(self, key, None)

    @classmethod
    def enum_for_field(cls, field_name):
        return enum_for_field(cls, field_name)

    @classmethod
    async def _resolve_reference_bulk(cls, model_instance, info):
        for field in info.parent_type.fields.values():
            if hasattr(field.type, "name"):
                field_type_name = field.type.name
                if (
                    field_type_name.endswith("Connection")
                    and field_type_name[:-10] == info.context.representation
                ):
                    return await field.resolve(model_instance, info)

    sort_enum = classmethod(sort_enum_for_object_type)

    sort_argument = classmethod(sort_argument_for_object_type)



"""
MIT License

Copyright (c) 2018 Mitchel Cabuloy

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import re
from typing import Union

import sqlalchemy
from graphene import ResolveInfo
from graphql import FieldNode, FragmentDefinitionNode

from alchql.registry import get_global_registry

_camel_to_snake_re = re.compile(
    r"((?!^[^A-Z]*)|\b[a-zA-Z][a-z\d]*)([A-Z]\d*[a-z]*|\d+)"
)


def camel_to_snake(name: str) -> str:
    name = _camel_to_snake_re.sub(r"\1_\2", name)
    return name.lower()


def collect_fields(node, fragments, cls_name: str = None):
    """Recursively collects fields from the AST

    Args:
        node (dict): A node in the AST
        fragments (dict): Fragment definitions

    Returns:
        A dict mapping each field found, along with their sub fields.

        {'name': {},
         'sentimentsPerLanguage': {'id': {},
                                   'name': {},
                                   'totalSentiments': {}},
         'slug': {}}
    """

    field = {}

    if node.get("selection_set"):
        for leaf in node["selection_set"]["selections"]:
            if leaf["kind"] == "field":
                field.update({leaf["name"]["value"]: collect_fields(leaf, fragments)})
            elif leaf["kind"] == "fragment_spread":
                field.update(
                    collect_fields(fragments[leaf["name"]["value"]], fragments)
                )
            elif (
                leaf["kind"] == "inline_fragment"
                and leaf["type_condition"].name.value == cls_name
            ):
                field.update(collect_fields(leaf, fragments))

    return field


def ast_to_dict(ast: Union[FieldNode, FragmentDefinitionNode]):
    result = {}
    for k in ast.keys:
        if k == "loc":
            continue
        value = getattr(ast, k)
        if k in {"name", "selection_set"}:
            value = ast_to_dict(value) if value else value
        if k == "selections":
            value = [ast_to_dict(i) for i in value]

        result[k] = value

    if hasattr(ast, "kind"):
        result["kind"] = ast.kind

    return result


def get_tree(info: ResolveInfo, cls_name: str = None):
    """A convenience function to call collect_fields with info

    Args:
        info (ResolveInfo)

    Returns:
        dict: Returned from collect_fields
    """

    fragments = {}
    node = ast_to_dict(info.field_nodes[0])

    for name, value in info.fragments.items():
        fragments[name] = ast_to_dict(value)

    return collect_fields(node, fragments, cls_name)


def get_fields(model, info: ResolveInfo, cls_name=None):
    tree = get_tree(info, cls_name)

    if "edges" in tree:
        tree = tree["edges"]
        if "node" in tree:
            tree = tree["node"]

    fields = []
    for key in tree.keys():
        if key == "__typename":
            continue

        model_names = {key, camel_to_snake(key)}
        for model_name in model_names:
            if hasattr(model, model_name):
                ex = getattr(model, model_name).expression
                break
        else:
            registry = get_global_registry()
            type_ = registry.get_type_for_model(model)
            for k, v in type_._meta.fields.items():
                if getattr(v, "name", None) in model_names:
                    ex = getattr(type_, k).model_field.expression
                    if k != ex.key:
                        ex = ex.label(k)
                    break
            else:
                raise Exception(f"Field not found: {key}")

        if hasattr(ex, "left") and hasattr(ex, "right"):
            if model.__table__ == ex.left.table:
                fields.append(ex.left)
            elif model.__table__ == ex.right.table:
                fields.append(ex.right)
        else:
            fields.append(ex)

    for pk in sqlalchemy.inspect(model).primary_key:
        fields.append(pk)
    fields = list(set(fields))

    return fields



import ast
import re
from datetime import datetime

from setuptools import find_packages, setup

_version_re = re.compile(r"__version__\s+=\s+(.*)")

with open("alchql/__init__.py", "rb") as f:
    version = str(
        ast.literal_eval(_version_re.search(f.read().decode("utf-8")).group(1))
    )
    version += f".{int(datetime.now().timestamp())}"

requirements = [
    # To keep things simple, we only support newer versions of Graphene
    "graphene>=3.0",
    "promise>=2.3",
    "SQLAlchemy>=1.4,<2",
    "aiodataloader == 0.2.1",
    "protobuf",
    "setuptools>=v49.0.0",
]

tests_require = [
    "pytest>=6.2.0,<7.0",
    "coverage[toml]",
    "sqlalchemy_utils>=0.37.0,<1.0",
    "pytest-benchmark>=3.4.0,<4.0",
    "pytest-asyncio>=0.17.2,<0.18",
    "aiosqlite>=0.17.0,<0.18",
]

setup(
    name="alchql",
    version=version,
    description="Graphene SQLAlchemy core integration",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/startupmillio/alchql",
    author="Yorsh Sergey",
    author_email="myrik260138@gmail.com",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords="api graphql protocol rest relay graphene",
    packages=find_packages(exclude=["tests"]),
    install_requires=requirements,
    extras_require={
        "test": tests_require,
    },
    tests_require=tests_require,
)



from typing import Optional, Type

from graphene import Connection, PageInfo
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Query

from .utils import (
    get_offset_with_default,
    offset_to_cursor,
)
from ..utils import filter_requested_fields_for_object


async def connection_from_query(
    query: Query,
    session: AsyncSession,
    args: Optional[dict] = None,
    list_length: int = 0,
    connection_type: Type[Connection] = Connection,
    page_info_type: Type[PageInfo] = PageInfo,
) -> Connection:
    """
    Given a slice (subset) of an array, returns a connection object for use in
    GraphQL.
    This function is similar to `connectionFromArray`, but is intended for use
    cases where you know the cardinality of the connection, consider it too large
    to materialize the entire array, and instead wish pass in a slice of the
    total result large enough to cover the range specified in `args`.
    """
    args = args or {}
    edge_type = connection_type.Edge

    before = args.get("before")
    after = args.get("after")
    first = args.get("first")
    last = args.get("last")

    before_offset = get_offset_with_default(before, list_length)
    after_offset = get_offset_with_default(after, -1)

    start_offset = max(after_offset, -1) + 1
    end_offset = min(before_offset, list_length)

    if isinstance(first, int):
        end_offset = min(end_offset, start_offset + first)
    if isinstance(last, int):
        start_offset = max(start_offset, end_offset - last)

    lower_bound = after_offset + 1 if after else 0
    upper_bound = before_offset if before else list_length

    _slice = query
    # If supplied slice is too large, trim it down before mapping over it.
    limit = first or last or end_offset
    if limit:
        _slice = _slice.limit(limit)
    if start_offset:
        _slice = _slice.offset(start_offset)

    edges = []

    for i, v in enumerate(await session.execute(_slice)):
        node_type = edge_type.node.type
        node_value = filter_requested_fields_for_object(dict(v), node_type)
        edge = edge_type(
            node=node_type(**node_value),
            cursor=offset_to_cursor(start_offset + i),
        )
        edges.append(edge)

    first_edge_cursor = edges[0].cursor if edges else None
    last_edge_cursor = edges[-1].cursor if edges else None

    return connection_type(
        edges=edges,
        page_info=page_info_type(
            start_cursor=first_edge_cursor,
            end_cursor=last_edge_cursor,
            has_previous_page=isinstance(last, int) and start_offset > lower_bound,
            has_next_page=isinstance(first, int)
            and len(edges) == first
            and end_offset < upper_bound,
        ),
    )



import logging
import re
from dataclasses import dataclass
from typing import Any, Optional, Type, Union

import graphene
import sqlalchemy as sa
from graphene import Field, ResolveInfo, Scalar
from graphene.types.objecttype import ObjectTypeMeta
from sqlalchemy import Table
from sqlalchemy.exc import ArgumentError
from sqlalchemy.orm import DeclarativeMeta, class_mapper, mapperlib, object_mapper
from sqlalchemy.orm.exc import UnmappedClassError, UnmappedInstanceError

from .gql_fields import get_fields


@dataclass
class FilterItem:
    field_type: Type[graphene.Field]
    filter_func: callable
    value_func: Optional[callable] = lambda x: x

    default_value: Any = None
    description: Optional[str] = None
    name: Optional[str] = None
    required: bool = False


@dataclass
class GlobalFilters:
    ID__EQ = "id__eq"
    ID__IN = "id__in"


def get_query(model: Type[DeclarativeMeta], info: ResolveInfo, cls_name=None):
    fields = None

    if info:
        try:
            fields = get_fields(model, info, cls_name)
        except Exception as e:
            logging.error(e)

    if not fields:
        fields = model.__table__.columns

    return sa.select(fields)


def is_mapped_class(cls):
    try:
        class_mapper(cls)
    except (ArgumentError, UnmappedClassError):
        return False
    else:
        return True


def is_mapped_instance(cls):
    try:
        object_mapper(cls)
    except (ArgumentError, UnmappedInstanceError):
        return False
    else:
        return True


def to_type_name(name):
    """Convert the given name to a GraphQL type name."""
    return "".join(part[:1].upper() + part[1:] for part in name.split("_"))


_re_enum_value_name_1 = re.compile(r"(.)([A-Z][a-z]+)")
_re_enum_value_name_2 = re.compile(r"([a-z\d])([A-Z])")


def to_enum_value_name(name):
    """Convert the given name to a GraphQL enum value name."""
    return _re_enum_value_name_2.sub(
        r"\1_\2", _re_enum_value_name_1.sub(r"\1_\2", name)
    ).upper()


class EnumValue(str):
    """String that has an additional value attached.

    This is used to attach SQLAlchemy model columns to Enum symbols.
    """

    def __new__(cls, s, value):
        return super().__new__(cls, s)

    def __init__(self, _s, value):
        super().__init__()
        self.value = value


def _deprecated_default_symbol_name(column_name, sort_asc):
    return column_name + ("_asc" if sort_asc else "_desc")


# unfortunately, we cannot use lru_cache because we still support Python 2
_deprecated_object_type_cache = {}


def _deprecated_object_type_for_model(cls, name):

    try:
        return _deprecated_object_type_cache[cls, name]
    except KeyError:
        from .types import SQLAlchemyObjectType

        obj_type_name = name or cls.__name__

        class ObjType(SQLAlchemyObjectType):
            class Meta:
                name = obj_type_name
                model = cls

        _deprecated_object_type_cache[cls, name] = ObjType
        return ObjType


def filter_value_to_python(value):
    """
    Turn the string `value` into a python object.
    >>> filter_value_to_python([1, 2, 3])
    [1, 2, 3]
    >>> filter_value_to_python(123)
    123
    >>> filter_value_to_python('true')
    True
    >>> filter_value_to_python('False')
    False
    >>> filter_value_to_python('null')
    >>> filter_value_to_python('None')
    >>> filter_value_to_python('Ø')
    u'O'
    """
    if isinstance(value, list):
        return value
    if isinstance(value, int):
        return value

    # Simple values
    if value in ["true", "True", True]:
        value = True
    elif value in ["false", "False", False]:
        value = False
    elif value in ("null", "none", "None", None):
        value = None

    return value


def filter_requested_fields_for_object(
    data: dict, conversion_type: Union[ObjectTypeMeta, object]
):
    if isinstance(conversion_type, ObjectTypeMeta):
        result = {}
        fields = conversion_type._meta.fields.keys()
        for key, value in data.items():
            if key in fields:
                result[key] = value
            else:
                attr = getattr(conversion_type, key, None)
                if attr and isinstance(attr, (Field, Scalar)):
                    result[key] = value

        return result
    if isinstance(conversion_type, Table):
        result = {}
        fields = conversion_type.columns.keys()
        for key, value in data.items():
            if key in fields:
                result[key] = value
            else:
                attr = getattr(conversion_type, key, None)
                if attr and isinstance(attr, (Field, Scalar)):
                    result[key] = value

        return result

    return data


def get_object_type_manual_fields(object_type):
    object_type_fields = {}
    for _name in dir(object_type):
        if _name.startswith("_"):
            continue

        attr = getattr(object_type, _name, None)
        if attr and isinstance(attr, (Field, Scalar)):
            if hasattr(attr, "kwargs") and attr.kwargs.get("name"):
                object_type_fields[attr.kwargs.get("name")] = attr
            else:
                object_type_fields[_name] = attr

    return object_type_fields


def table_to_class(table: Table) -> DeclarativeMeta:
    for mapper_registry in mapperlib._all_registries():
        for mapper in mapper_registry.mappers:
            if table in mapper.tables:
                return mapper.entity



import re
from inspect import isawaitable
from typing import Sequence, Type, Union

import sqlalchemy as sa
from graphene import ResolveInfo
from sqlalchemy.orm import DeclarativeMeta, Mapper

from alchql.loader_fk import (
    generate_loader_by_foreign_key,
    generate_loader_by_relationship,
)


class LoaderMiddleware:
    def __init__(self, models: Sequence[Union[Mapper, Type[DeclarativeMeta]]]):
        self.loaders = {}
        for model in models:
            if isinstance(model, Mapper):
                model = model.entity

            inspected_model = sa.inspect(model)
            for fk in inspected_model.mapped_table.foreign_keys:
                key = (
                    fk.constraint.table,
                    fk.constraint.referred_table,
                    re.sub(r"_(?:id|pk)$", "", fk.parent.key),
                )
                self.loaders[key] = generate_loader_by_foreign_key(fk)

                key = (
                    fk.constraint.referred_table,
                    fk.constraint.table,
                    str(fk.constraint.table.fullname),
                )
                self.loaders[key] = generate_loader_by_foreign_key(fk, reverse=True)

            for relationship in inspected_model.relationships.values():
                key = (
                    relationship.parent.entity,
                    relationship.mapper.entity,
                    relationship.key,
                )
                self.loaders[key] = generate_loader_by_relationship(relationship)

    async def resolve(self, next_, root, info: ResolveInfo, **args):
        if root is None:
            session = info.context.session

            info.context.loaders = {k: v(session) for k, v in self.loaders.items()}

        result = next_(root, info, **args)
        if isawaitable(result):
            return await result
        return result
